import torch.nn as nn


def ignore_warning(category=None):
    import warnings
    if category is not None:
        warnings.filterwarnings("ignore", category=category)
    else:
        warnings.filterwarnings('ignore')


def set_plot_formats(formats='svg'):
    from matplotlib_inline import backend_inline
    backend_inline.set_matplotlib_formats(formats)


def try_use_device(cuda=True, tensor=True):
    import torch
    if cuda and torch.cuda.is_available():
        device = torch.device("cuda")
        type = torch.cuda.FloatTensor
    else:
        device = torch.device("cpu")
        type = torch.FloatTensor

    if tensor:
        torch.set_default_tensor_type(type)

    torch.set_default_device(device)
    return device


def load_model(model, filename) -> nn.Module:
    import torch
    model.load_state_dict(torch.load(filename))
    model.eval()
    return model


def save_model(model, filename):
    import os
    import torch

    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    torch.save(model.state_dict(), filename)


def save_module(model, filename):
    from .attrdict import attrdict
    import pickle
    import os
    attr = attrdict()
    attr.classname = type(model)
    attr.state_dict = model.state_dict()

    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(filename, 'wb') as file:
        file.write(pickle.dumps(attr))


def load_module(filename) -> nn.Module:
    import pickle
    with open(filename, 'rb') as file:
        attr = pickle.loads(file.read())

    model = attr.classname()
    model.load_state_dict(attr.state_dict)
    model.eval()
    return model


def save_pickle(model, filename):

    import pickle
    import os

    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(filename, 'wb') as file:
        file.write(pickle.dumps(model))


def load_pickle(filename) -> nn.Module:
    import pickle
    with open(filename, 'rb') as file:
        model = pickle.loads(file.read())

    if hasattr(model, 'eval') and callable(model.eval):
        model.eval()

    return model


def metrics(y_true, y_pred):
    from .attrdict import attrdict
    from sklearn import metrics as m
    import math

    scores = attrdict()
    scores.MAPE = m.mean_absolute_percentage_error(y_true, y_pred)
    scores.MSE = m.mean_squared_error(y_true, y_pred)
    scores.RNSE = math.sqrt(scores.MSE)
    scores.MAE = m.mean_absolute_error(y_true, y_pred)
    scores.EVS = m.explained_variance_score(y_true, y_pred)
    scores.R2 = m.r2_score(y_true, y_pred)

    return scores


def naive_fit(model, dataset, criterion, optimizer, epoch, batch_size, device, shuffle, progress=True, step_callback=None):
    import torch
    from tqdm import tqdm

    dataloader = torch.utils.data.DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        drop_last=True,
        generator=torch.Generator(device=device),
        shuffle=shuffle,
    )

    batch = len(dataloader)

    losses = []

    def updater(x, t):
        y = model.forward(x)
        loss = criterion(y, t)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        return loss

    if step_callback is None:
        step_callback = updater

    with tqdm(total=epoch * batch) as bar:
        for e in range(epoch):
            for i, (x, t) in enumerate(dataloader):

                loss = step_callback(x, t)
                losses.append(loss.item())

                if progress:
                    bar.set_description(f"({e + 1:02}/{epoch}) | ({i + 1:02}/{batch})")
                    bar.update()
                    bar.set_postfix(loss=f"{loss.item():0.6}")

    return losses


def naive_test(models, dataset, batch_size, device, progress=True):
    import torch
    from tqdm import tqdm

    dataloader = torch.utils.data.DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        drop_last=True,
        generator=torch.Generator(device=device),
        shuffle=False,
    )

    if progress:
        dataloader = tqdm(dataloader)

    preds = []
    for model in models:
        model.eval()
        preds.append([])

    real = []

    for x, t in dataloader:
        real.append(t)
        for i, model in enumerate(models):
            y = model.forward(x)
            preds[i].append(y)

    real = torch.cat(real)
    for i, pred in enumerate(preds):
        preds[i] = torch.cat(pred)

    return real, preds



# coding=utf-8
import time
import json
import unittest

from test_base import BaseTestCase


class TestCase(BaseTestCase):

    def test_utils(self):
        import tone
        utils = tone.utils
        return utils

if __name__ == '__main__':
    TestCase.main()



# coding=utf-8
import sys
from io import StringIO

from test_base import BaseTestCase
import logging


class TestCase(BaseTestCase):

    def setUp(self):
        pass

    def test_logger(self):
        stderr = sys.stderr
        try:
            sys.stderr = StringIO()
            from tone import utils
            logger = utils.get_logger()
            logger.debug('hello world')
            self.assertTrue('hello world' in sys.stderr.getvalue())
        finally:
            sys.stderr = stderr


if __name__ == '__main__':
    TestCase.main()



# coding=utf-8
import unittest
import importlib


modules = [
    'test_utils_attrdict',
]


def main():
    suite = unittest.TestSuite()
    for name in modules:
        module = importlib.import_module(name)
        suite.addTest(unittest.makeSuite(module.TestCase))

    # failfast 如果失败则推出
    runner = unittest.TextTestRunner(failfast=True)
    runner.run(suite)


if __name__ == '__main__':
    main()




def autoreload(module=None):
    import types

    try:
        from IPython import get_ipython
    except ImportError:
        print('Run in IPython')
        return

    # https://ipython.readthedocs.io/en/stable/config/extensions/autoreload.html
    if module is None:
        module = __name__
    elif isinstance(module, types.ModuleType):
        module = module.__name__

    ipy = get_ipython()
    if 'IPython.extensions.autoreload' not in ipy.extension_manager.loaded:
        ipy.run_line_magic('load_ext', 'autoreload')

    ipy.run_line_magic('autoreload', '1')
    ipy.run_line_magic('aimport', module)


class StopExecution(Exception):

    def _render_traceback_(self):
        pass



# coding=utf-8
import os
import sys
import unittest

dirname = os.path.dirname(os.path.abspath(__file__))
project = os.path.abspath(os.path.join(dirname, '..'))

if project not in sys.path:
    sys.path.insert(0, project)


class BaseTestCase(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @staticmethod
    def main():
        return unittest.main()



# coding=utf-8
import time
import json
import unittest

from test_base import BaseTestCase

from tone import utils
import torch.nn as nn


class Model(nn.Module):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.model = nn.Sequential(
            nn.Linear(16, 32),
            nn.Linear(32, 1),
        )


class TestCase(BaseTestCase):

    def test_metrics(self):
        from tone.utils.learning import metrics
        import numpy as np

        real = np.linspace(0, 10, 100)
        pred = real + np.random.randn(real.shape[0]) * 0.01
        scores = metrics(real, pred)

    def test_ignore_warning(self):
        import tone
        import warnings
        tone.utils.learning.ignore_warning()
        warnings.warn("test warning..")

    def test_save_load_module(self):
        import os
        from tone.utils import learning

        model = Model()
        filename = './test_model.pt'
        learning.save_module(model, filename)
        model = learning.load_module(filename)
        self.assertEqual(type(model), Model)
        os.remove(filename)

    def test_save_load_pickle(self):
        import os
        from tone.utils import learning

        model = Model()
        filename = './test_model.pt'
        learning.save_pickle(model, filename)
        model = learning.load_pickle(filename)
        self.assertEqual(type(model), Model)
        os.remove(filename)


if __name__ == '__main__':
    TestCase.main()



# coding=utf-8
import time
import json
import unittest

from test_base import BaseTestCase

from tone import utils


class TestCase(BaseTestCase):

    def test_attrdict(self):
        from tone.utils.attrdict import attrdict
        data = attrdict()
        self.assertFalse(data)
        with self.assertRaises(AttributeError):
            data.key

        data.key = 1
        self.assertEqual(data.key, 1)

        self.assertEqual(json.dumps(data), '{"key": 1}')

    def test_defaultattrdict(self):
        from tone.utils.attrdict import defaultattrdict
        data = defaultattrdict()
        self.assertFalse(data)
        self.assertFalse(data.data)
        self.assertFalse(data.data.data.data)

        data.data.data.data = 1
        self.assertTrue(callable(data.items))
        self.assertTrue(callable(data.values))

        self.assertEqual(json.dumps(data), '{"data": {"data": {"data": 1}}}')

    def test_loads(self):
        from tone.utils.attrdict import attrdict

        data = {'key': 123}
        data = attrdict.loads(data)
        self.assertIsInstance(data, attrdict)
        self.assertEqual(data.key, 123)

        data = {
            'key': [
                {
                    'key': 123,
                },
                [
                    {
                        'key': 123,
                    },
                    1234
                ]
            ]
        }
        result = attrdict.loads(data)
        self.assertEqual(str(data), str(result))

        self.assertIsInstance(data, dict)
        self.assertNotIsInstance(data, attrdict)
        self.assertIsInstance(result, dict)
        self.assertIsInstance(result, attrdict)
        self.assertIsInstance(result.key, list)

        self.assertEqual(result.key[0].key, 123)

    def test_json_loads(self):
        json = '{"data":{"data":1},"datas":[{"data":1}]}'
        from tone.utils.attrdict import defaultattrdict
        from tone.utils.attrdict import attrdict

        data = attrdict.json_loads(json)
        self.assertTrue(data)
        self.assertIsInstance(data, attrdict)

        data = defaultattrdict.json_loads(json)
        self.assertTrue(data)
        self.assertIsInstance(data, defaultattrdict)

        json = '[{"data":1},{"data":1}]'
        data = defaultattrdict.json_loads(json)
        self.assertTrue(data)
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 2)

    def test_performance(self):
        from tone.utils.attrdict import defaultattrdict
        from tone.utils.attrdict import attrdict
        count = 100000

        def test_dict():
            data = {}

            start = time.time()
            for _ in range(count):
                data['key'] = 1
                assert data['key'] == 1

            end = time.time()
            print('test dict', end - start)

        def test_attrdict():
            data = attrdict()
            start = time.time()
            for _ in range(count):
                data.key = 1
                assert data.key == 1

            end = time.time()
            print('test attrdict', end - start)

        def test_defaultattrdict_attr():
            data = defaultattrdict()
            start = time.time()
            for _ in range(count):
                data.key = 1
                assert data.key == 1

            end = time.time()
            print('test defaultattrdict attr', end - start)

        def test_defaultattrdict_item():
            data = defaultattrdict()
            start = time.time()
            for _ in range(count):
                data['key'] = 1
                assert data['key'] == 1

            end = time.time()
            print('test defaultattrdict item', end - start)

        test_dict()
        test_attrdict()
        test_defaultattrdict_attr()
        test_defaultattrdict_item()


if __name__ == '__main__':
    TestCase.main()



# coding=utf-8
import logging
import logging.config


def get_basic_logger(name='tone'):
    '''
    Get logger for convenient method
    '''

    logger = logging.getLogger(name)
    if len(logger.handlers) > 0:
        return logger

    config = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'verbose': {
                'format': '[%(asctime)s] [%(module)s] [%(lineno)d] [%(levelname)s] | %(message)s'
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'verbose',
                "level": "DEBUG",
            },
        },
        'loggers': {
            name: {
                'handlers': ['console', ],
                'level': 'DEBUG',
                'propagate': True,
            },
        },
    }
    logging.config.dictConfig(config)
    logger = logging.getLogger(name)
    return logger


def get_logger(name='tone'):
    return get_basic_logger(name)
